// This is an example of how to drive the W5XD Multi+ keyer
// It opens a COMM port and sends commands to the keyer
//
// Copyright (c) 2001 by Wayne E. Wright, W5XD, Round Rock, Texas
// All rights reserved.
//
// Permission is GRANTED to copy/modify/use the software routines in this file.
// I simply require that, should you incorporate this into your product,
// that you credit W5XD in your own About Box, and that you inform
// me at w5xd@writelog.com that your product supports the W5XD multi+ keyer.
//
//
// NOTE: these routines ONLY support the W5XD multi+ keyer firmware versions
// 10.15 and later which require the 28 pin microprocessor. These routines 
// DO NOT support the old W5XD outboard keyer on a 16 pin DIP processor.

#include <windows.h>
#include <stdio.h>

#define NUM_CW_SPEEDS 16
#define NUM_CW_WEIGHTS 5

/* keyer command bytes */
//anything below 0xC0 is a morse character
// 0x00 is special case--it is a word space
#define PCM_DATA   0xE0		// load 5 bits of keyer's 3 byte holding buffer
#define PCM_LATCH  0xC1		// head phone latch--keyer holds headphones rcv radio until xmit ends
#define PCM_ABORT  0xC0		// keyer clears its buffer, sets CW off
#define PCM_XFRSP  0xC2		// transfer speed
#define PCM_XFRWT  0xC3		// transfer weight
#define PCM_DEFSPE 0XC4		// transfer speed (handled in order of data bytes
#define PCM_UNLOCK 0xC5		// use of paddles locks the keyer--it will not accept cw commands until unlocked
#define PCM_XFRATO 0XC6		// transfer keyer status bits
#define PCM_ANTRLY 0XC7		// latch keyer's antenna relay outputs
#define PCM_PTTON  0XC8		// tell keyer to turn PTT ON
#define PCM_PTTOFF	0XC9	// tell keyer to turn PTT OFF
#define PCM_XFSTONE 0XCA	// transfer side tone parameters
#define PCM_XFWAVEST 0xCB	// transfer wave status bits
#define PCM_DELETE 0xCC		// keyer deletes the newest morse char in its buffer, if there is one
#define PCM_PTTDELAY 0XCD	// NOT IN FIRMWARE 10.19 AND EARLIER. set keyer's PTT Delay setting for paddle input.

#define MAX_MULTI_MSEC 50

#define SPACE_CHAR 0

HANDLE CommChannel;

// Some PC hardware has no DSR circuit.
// The keyer sends the same signal on CD as it does on DSR,
// so if both are connected, it makes no difference how this flag is set.
static int UseCDInsteadOfDSR;


// The keyer has these five status bits that can be set from the PC

// Operator preference for whether paddle-sent CW dis-allows spaces bigger than
// a letter but smaller than a word.
static int AutoLtrSpace = 0;	//0x10

// The keyer puts its output on either L or R
static int RadioLR = 1;			//0x4

// The keyer controls audio relays. Set this bit ON to split the headphones
static int HpSplit = 0;			//0x8

// The keyer has a side tone separately controlled for paddles and machine-sent
static int StonePaddles = 1;	//0x2
static int StoneMachineSent = 1;	//0x1


// ASCII to keyer conversion table.
// An ASCII character converts to a keyer character according to Morse code.
// Morse is represented in this table as 0 bit is a dot, 1 bit is a dash,
// along with one extra leading 1 bit to mark the end of the Morse character
// The LSB in this table is the first dot/dash to be sent.
const unsigned char asc_to_m[128] =
	{       /*read the binary right to left, 0's are dots. ignore leading 1*/
	0, 0, 0, 0, 0, 0, 0, 0, 0,
	052/*0b101010 tab is AR*/, 0,0,0, SPACE_CHAR, /*CR becomes space*/
	0,0,0,0,0,0,0,0,0,0,0,0,0,
	0150/*b1101000 ESC is SK*/, 0,0,0,0,
	SPACE_CHAR,     /*special flag for space*/
	0, /*!*/
   0, /* " */
   055, /* # */
   0150, /* $ */ 0, /* % */ 0, /* & */
	052, /*AR= ' */ 0, /* ( */  0, /* ) */ 0, /* * */ 032, /* + = AA */ 
	0163, /*0xb1110011 , */
	061,    /*0xb110001 - */
	0152,   /*0xb1101010 . */
	051,    /*0xb101001 / */
	077,            /*0xb111111 0 */
	076,                    /*0xb111110 1 */
	074,    070, 060, 040,  /* 2, 3, 4, 5 */
	041,    043, 047, 057,  /* 6, 7 , 8, 9 */
	0170,   /*:*/ 
	0125, /*0b1010101 ; */ 0, 0, 0,
	0114,   /*0b1001100 ? */        052, /* @ */
	06, /*110*/     /*A*/
	021, /*0b10001*/        /*B*/
	025, /*0b10101*/        /*C*/
	011, /*0b1001*/         /*D*/
	2,      /*E*/
	024,    /*0b10100 F*/
	013, 020, 4, 036, 015, 022, 7, 5, 017, 026, 033, 012, 010, 3, 014,
	030, 016, 031, 035, 023,
	0150,   /*0b1101000 [ is SK*/
	0,
	042,    /*0b100010 ] is AS */
	1,		/* dot space <NOT SUPPORTED IN 10.19 AND EARLIER FIRMWARE>*/
	0,0,
	06, /*110*/     /*a*/
	021, /*0b10001*/        /*b*/
	025, /*0b10101*/        /*c*/
	011, /*0b1001*/         /*d*/
	2,      /*e*/
	024,    /*0b10100 f*/
	013, 020, 4, 036, 015, 022, 7, 5, 017, 026, 033, 012, 010, 3, 014,
	030, 016, 031, 035, 023,
    0,0,0,0,
	0200    /*DELETE*/
	};

// The keyer takes 2 parameter to set its speed.
// This is the length of a dot in units of 512 usec
// This table corresponds to:
//    7, 12, 13, 14, 15, 16, 18, 20, 22, 25, 28, 32, 35, 40, 45, 50 WPM
static int DotTimeKeyer[NUM_CW_SPEEDS] =  /* Length of a dot in 512 usec */
{
	  156   ,
	  156,
	  156,
	  156,
	  156,
	  146,
	  130,
	  117,
	 107,
	 94,
	 84,
	 73,
	 67,
	 59,
	 52,
	 47
};


static int LtrTimeKeyer[NUM_CW_SPEEDS] =
{
       1406,      /* length of a letter space in msec. word space is 2.5 of these*/
       586,
       490,
       397,
       335,
       293,
       260,
      234,
       213,
      188,
      167,
      146,
     134,
      117,
	104,
	94
};

static int KeyerWeights[NUM_CW_WEIGHTS] =
   {
      -15  ,
      -5 ,
      0,
      5 ,
      15, /* UNITS ARE 512 usec clock ticks*/
   };

// Useful conversions of WPM to dot length
static int DotLen(int Wpm,
         int TimeBase /* microseconds */)	// use 512 for this keyer
{
   if (Wpm >= 15)
      return (60L * 1000000L)/((long)Wpm * 50L * (long)TimeBase);
   else
      return (60L * 1000000L)/( 15L * 50L * (long) TimeBase);
}

static int LtrLen(int Wpm,
                  int TimeBase) /* microseconds*/
{
   if (Wpm >= 15)
      return (2 * DotLen(Wpm, TimeBase));
   else
      return (60000000L - (36L * 10L * DotLen(15, 10) * Wpm)) /
            ((long) Wpm * 7 * (long) TimeBase);
   // There are 36 dotlens in a word, and they are sent at 15 wpm
   //There are 7 ltrlen's in a word, and they are sent at requested speed
}


static int GetKeyerCD()
{
	DWORD CommStat;
    GetCommModemStatus(CommChannel, &CommStat);
    return ((CommStat & MS_RLSD_ON ) ? 1 : 0);
}

static int GetKeyerDSR()
{
    DWORD CommStat;

    GetCommModemStatus(CommChannel, &CommStat);
    return ((CommStat & MS_DSR_ON) ? 1 : 0);
}

static int GetKeyerCTS()
{
    DWORD CommStat;
    GetCommModemStatus(CommChannel, &CommStat);
    return ((CommStat & MS_CTS_ON) ? 1 : 0);
}

static void SignalMultiTimeout()
{
	// The keyer is not responding.. do something...
	printf("Oops, Keyer not talking!\n");
}

// The keyer accepts commands as the PC does a series
// of status changes on RTS/DTR. At the same time as
// the PC sends a byte to the keyer, the keyer is sending
// one back to the PC...
static int ExchangeStateSer(int OutState, int NoWait)
{
	static int CommCommand[4][3] =
	{
		{SETRTS, SETDTR, 0},	//last column is parity
		{SETRTS, CLRDTR, 1},
		{CLRRTS, SETDTR, 1},
		{CLRRTS, CLRDTR, 2},
	};

	DWORD Timer;
	int DSR;
	int CTS;

	EscapeCommFunction(CommChannel, CommCommand[OutState][0]);
	EscapeCommFunction(CommChannel, CommCommand[OutState][1]);

	if (NoWait)
		return -1;
	Timer = GetTickCount();
	for (;;)
	{
		DSR = UseCDInsteadOfDSR ?
			(GetKeyerCD() ? 0 : 1) :
				(GetKeyerDSR() ? 0 : 1);
		CTS = GetKeyerCTS() ? 0 : 1; 
		if ((DSR + CTS) == CommCommand[OutState][2])
			break;
		if ((GetTickCount() - Timer) > MAX_MULTI_MSEC)
		{
			SignalMultiTimeout();
			return -1;
		}
	}
	return (DSR << 1) | CTS;
}

static int KeyerFifoOK()
{
	int State;
	if (ExchangeStateSer(0, 0) < 0)
		return 0;
	State = ExchangeStateSer(2, 0);
	if (State < 0)
		return 0;
	if (ExchangeStateSer(3, 0) < 0)
		return 0;
	State = ExchangeStateSer(2, 0);
	if (State < 0)
		return 0;
	return !(State & 1); 
}

// Note we only call ExchangeState twice, so this is 
// several times quicker than sending a whole byte
static int KeyerXmitInProgress()
{
	int State;
	if (ExchangeStateSer(0, 0) < 0)
		return 0;
	State = ExchangeStateSer(2, 0);
	if (State < 0)
		return 0;
	return (State & 1);
}

static void LoadKeyer(char KeyerCommand)
{
    int i;
	int State;
	int IsSpeed = !(KeyerCommand & 0x80);
	unsigned char Speed = 0;

	if (ExchangeStateSer(0, 0) < 0)
		return;

    for (i = 8; ;)
    {
		State = ExchangeStateSer( (KeyerCommand & 0x80) ? 2 : 1, 0);
		if (State < 0)
			return;
		State &= 1;
		Speed |= State;
		i -= 1;
		if (i == 0)
		   break;
		if (ExchangeStateSer(3, 0) < 0)
			return;
		Speed <<= 1;
		KeyerCommand <<= 1;
    }
	ExchangeStateSer(0, 1);
#if 0
	// The keyer has a knob. If Speed is different that 
	// what PC last set, then we can do something about that,
	// like display the knob speed
	if (IsSpeed && Speed && (Speed != DotTimeKeyer[CurSpeed]))
	{
		
	}
#endif
}

static void SetKeyerParams(void)
{
	char ToLoad = 0;
	if (AutoLtrSpace)
		ToLoad |= 0x10;
	if (RadioLR)
		ToLoad |= 0x4;
	if (HpSplit)
		ToLoad |= 0x8;
	if (StonePaddles)
		ToLoad |= 0x2;
	if (StoneMachineSent)
		ToLoad |= 0x1;

	LoadKeyer((char)(PCM_DATA | (ToLoad & 0x1F)));
    LoadKeyer((char)(PCM_XFRATO));
}


static void SetKeyerSideTone()
{
	unsigned char	CyclesOn = 1;
	unsigned char	CyclesOff = 4;
	unsigned long	PatternToLoad = CyclesOff & 0xFF | ((CyclesOn << 8) & 0xFF00u);
	//there are 16 bits to load. get 5 at a time. 
	int i = 4;
	PatternToLoad <<= 4;		//we're loading a total of 20 bits
	while (i)
	{
		LoadKeyer((char)(PCM_DATA | (PatternToLoad & 0x1Fu)));
		PatternToLoad >>= 5;
		i -= 1;
	}

	LoadKeyer((char)PCM_XFSTONE);
}

static void SetKeyerPttDelay()
{
	unsigned long	PttDelay = 4;	// units of msec
	unsigned long	PttIdleCount = 500;	// units of msec
	unsigned long	BitsToSend;
	int i = 3;

#if 0
	if (don't want paddles to set PTT)
		PttDelay = 0;
#endif

	//there are 8 bits to load. get 5 at a time. Must send 10, so shift 2
	PttDelay <<= 1;		//convert msec to 512 usec
	if (PttDelay > 255)
		PttDelay = 255;	// = 128 msec max
	PttIdleCount >>= 4;	//convert msec to units of 16 msec
	if (PttIdleCount > 127)
		PttIdleCount = 127;	// = 2 seconds max
	BitsToSend = ((PttDelay << 7) & 0x7F80) | (PttIdleCount & 0x7F);
	while (i)
	{
		LoadKeyer((char)(PCM_DATA | (BitsToSend & 0x1Fu)));
		BitsToSend >>= 5;
		i -= 1;
	}

	LoadKeyer((char)PCM_PTTDELAY);
}

static void SetKeyerSpeed(int SpeedIndex)
{
#ifdef _DEBUG
	char	TimerMsg[256];
	DWORD	Time = (DWORD) - (long)GetTickCount();
#endif
   unsigned long  TotalBytes;
   int i;

   if (SpeedIndex < 0)
	   SpeedIndex = 0;
   if (SpeedIndex >= NUM_CW_SPEEDS)
	   SpeedIndex = NUM_CW_SPEEDS - 1;
   


   TotalBytes =    ((unsigned long)DotTimeKeyer[SpeedIndex] << 16      ) &0xFF0000u |
                   ((unsigned long)LtrTimeKeyer[SpeedIndex] ) & 0xFFu |
                   ((unsigned long)LtrTimeKeyer[SpeedIndex] ) & 0xFF00u ;
   TotalBytes <<= 1;  //The bottom bit gets lost in the transfer
   for (i = 0; i < 5; i += 1)
   {
      LoadKeyer((char)(PCM_DATA | (TotalBytes & 0x1F)));
      //The keyers PCSPEED variables TOP 5 bits are now the BOTTOM
      //five bits of TotalBytes. The lower three bits have been
      //shifted into PCSPEED+1 and +2

      TotalBytes >>= 5;
   }
   LoadKeyer((char)PCM_XFRSP);
#ifdef _DEBUG
	Time += GetTickCount();
	wsprintf(TimerMsg, "SetKeyerSpeed: %d\n", Time);
	OutputDebugString(TimerMsg);
#endif
}
static void SetKeyerWeight(int WeightIndex)
{
   short WeightMask;
   int i;

   if (WeightIndex < 0)
	   WeightIndex = 0;
   if (WeightIndex >= NUM_CW_WEIGHTS)
	   WeightIndex = NUM_CW_WEIGHTS - 1;
   WeightMask = KeyerWeights[WeightIndex] << 2;

   for (i = 0; i<2; i += 1)
   {
      LoadKeyer((char)(PCM_DATA | (WeightMask & 0x1F)));
      WeightMask >>= 5;
   }
   LoadKeyer((char)PCM_XFRWT);
}

// Set the antenna relay output for one radio or the other
static void SetAntRelay(int Radio, int Band)
{
	if (Band >= 0)
	{
		Band %= 15;		//There are 15 valid settings (all zeros is not used)
					//range is 0 to 14
		Band += 1;	//range is 1 to 15
		if (Radio)
		{
		}
		else
			Band |= 0x10;
		LoadKeyer((char)(PCM_DATA | Band));
		LoadKeyer((char)PCM_ANTRLY);
	}
}

static void LoadDeferredSpeed(int Index)
{  
	// This command is loaded in the keyer's FIFO, so have to check it first
	while (!KeyerFifoOK())
	{
		Sleep(300);	//Keyer WILL OVERFLOW if PC sends it a character when its FIFO is full
		printf("Waiting for keyer...\n");
	}

   unsigned short SpeedBin;
   int i;
   if (Index < 0)
      Index = 0;
   if (Index >= NUM_CW_SPEEDS)
      Index = NUM_CW_SPEEDS - 1;

   SpeedBin = DotTimeKeyer[Index];
   if (SpeedBin >= 0x80)
      return;     /* can't do it if takes the eighth bit */
   /* There's no shift of SpeedBin here because
   ** its bottom 10 bits shifted into an 8 bit byte drop
   ** the bottom 2 bits of SpeedBin, but that's exactly
   ** how much the PCM_DEFSPE command needs.
   */
   SpeedBin |= 0xC0 << 2;  /* set the speed byte indicator */
   for (i = 0; i<2; i += 1)
   {
      LoadKeyer((char)(PCM_DATA | (SpeedBin & 0x1F)));
      SpeedBin >>= 5;
   }
   LoadKeyer((char)PCM_DEFSPE);

}



static void SendASCII(char c)
{
	// This command is loaded in the keyer's FIFO, so have to check it first
	while (!KeyerFifoOK())
	{
		Sleep(300);	//Keyer WILL OVERFLOW if PC sends it a character when its FIFO is full
		printf("Waiting for keyer...\n");
	}
	LoadKeyer(asc_to_m[c & 0x7F]);
}


// Here are just some examples of how to drive the keyer....
static void TestKeyerCommands()
{
	// Make sure its initialized
	SetKeyerParams();
	SetKeyerSideTone();
	SetKeyerSpeed(11);
	SetKeyerWeight(2);

	char *p = "TEST DE W5XD";

	printf("Sending TEST DE W5XD\n");
	while (*p)
	{
		SendASCII(*p++);
	}

	while (KeyerXmitInProgress())
	{
		printf("Waiting for keyer to finish...\n");
		Sleep(300);
	}

	// We can change the speed in the middle of the message by inserting LoadDeferredSpeed along the way

	printf("Now Sending TEST <TEST> TEST\n");

	p = "TEST ";
	while (*p)
		SendASCII(*p++);
	LoadDeferredSpeed(14);
	p = "TEST ";
	while (*p)
		SendASCII(*p++);

	LoadDeferredSpeed(11);
	p = "TEST ";
	while (*p)
		SendASCII(*p++);



	while (KeyerXmitInProgress())
	{
		printf("Waiting for keyer to finish...\n");
		Sleep(300);
	}
		// keyer won't send final character if we don't hold DTR/DTS
		// long enough, and exiting this program resets them
}

// entry point
// Get a comm port is all we do to start...
int main()
{
	static char CommPort[] = "COM4";

	// note that this program does not bother to set any comm port parameters
	// as it only uses the modem control lines and not the data lines on the
	// comm port. 
	// Rig control can be accomplished on the same port at the same time
	// as the keyer....
	CommChannel = CreateFile(CommPort,
									GENERIC_READ|GENERIC_WRITE,
									0, 0, OPEN_EXISTING, 
									FILE_ATTRIBUTE_NORMAL | FILE_FLAG_OVERLAPPED, 
									0);

	if (CommChannel != INVALID_HANDLE_VALUE)
	{
		TestKeyerCommands();
		CloseHandle(CommChannel);
	}
	else
	{
		printf("Failed to open %s\n", CommPort);
	}

	return 0;
}